/*
 Dahdi layer for Quasar E1 interface board driver

 Copyright (C) 2006-2007 Parabel Ltd.
 Author: Mikhail Khelik <mixel.net@gmail.com>

 Copyright (C) 2008-2009 Parabel Ltd.
 Author: Pavel Selivanov <pavel[AT]parabel.ru>

 * Parabel Ltd. reserve the right to distribute project or it's parts
 under the alternative conditions.

 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software Foundation;
 either version 2 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with this program;
 if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/completion.h>
#include <linux/list.h>
#include <linux/types.h>
#include <linux/interrupt.h>

#include "dahdi/kernel.h"

#include "dver.h"
#include "pmc_e1_conf.h"
#include "pmc_e1.h"
#include "e12zap.h"

/* Definitions */
#define QUASAR_MAX_SPANS 8
/* Single DMA buffer size for all SPANS on board. */
#define QUASAR_BASE_SIZE (DAHDI_CHUNKSIZE * 32 * QUASAR_MAX_SPANS)

#if (DAHDI_CHUNKSIZE != 8)
#error Sorry, Quasar does not support chunksize != 8.
#endif

struct quasar_span {
	struct dahdi_span span;
	struct dahdi_chan _chans[31];	/* omit 0 timeslot */
	struct dahdi_chan *chans[31];	/* omit 0 timeslot */
	int spannumber;			/* span number in card */
	int syncprior;			/* span sync priority within board */
	void *pvt;
	int clear_counters;
	unsigned long frok;                                 /* Frame is stable frok ticks */
	unsigned char ec_chunk1[31][DAHDI_CHUNKSIZE];	/* first EC chunk buffer */
	unsigned char ec_chunk2[31][DAHDI_CHUNKSIZE];	/* second EC chunk buffer */
};

struct quasar_dpmc {
	volatile unsigned int *writechunk;	/* Double-word aligned write memory */
	volatile unsigned int *readchunk;	/* Double-word aligned read memory */
	dma_addr_t readdma;
	dma_addr_t writedma;
	struct quasar_span *spans[QUASAR_MAX_SPANS];
	unsigned char num_spans;	/* how many spans this card has */
	unsigned char qn_span;		/* For fast check during the interrupt*/
	char registered;
	void *pvt;
	u32 ovfl_counter;		/* keep old value of overflow counter, used in e12zap_delivery */
	char check;			/* used for checking interrupt misses */
	int intcount;
	int irqmisses;        		/* digium defined it as signed... */
#ifdef E12ZAP_TASKLET
	int taskletrun;
	int taskletsched;
	char taskletpending;
	int taskletexec;
	int txerrors;
	struct tasklet_struct e12zap_tlet;
#endif
};

/* Function prototypes  */
void e12zap_proc(struct quasar_dpmc *dpmc);

struct quasar_span *qspan_from_dspan(struct dahdi_span *span)
{
        return container_of(span, struct quasar_span, span);
}

void e12zap_find_master(struct quasar_dpmc *dpmc)
{
	struct quasar_pmc *pmc = dpmc->pvt;
	int i, syncsrc, master;

	master = pmc->bar0[0x02] & 0x07;
	/* Can be used for debug later.
	 syncsrc = dpmc->spans[master]->span.spanno;
	 ts = dpmc->spans[master]; */

	if (pmc->bar0[0x02] & (MODE1_HOLD|MODE1_FREE))
		syncsrc = 0; /* Using internal sync source for CT-BUS */

	for (i=0; i < dpmc->num_spans; ++i) {
		dpmc->spans[i]->span.syncsrc = syncsrc;
	}
}

#ifdef HAVE_SPAN_COUNT
# define DAHDI_COUNT(a,b,c) a.c
#else
# define DAHDI_COUNT(a,b,c) a.b
#endif //HAVE_DAHDI_COUNT

void e12zap_check_alarm(struct quasar_dpmc *dpmc)
{
	int i;
        unsigned long reg;
	struct quasar_span *qspan;
	struct quasar_pmc *pmc = dpmc->pvt;

	for (i=0; i < dpmc->num_spans; ++i) {
		qspan = dpmc->spans[i];

		if (framer_checkbit(pmc, i, STAT0_RAIS_HP)) {
			qspan->span.alarms |= DAHDI_ALARM_YELLOW;
		} else {
			qspan->span.alarms &= ~DAHDI_ALARM_YELLOW;
		}
		if (framer_checkbit(pmc, i, STAT0_LOF_HP)) {
			qspan->span.alarms |= DAHDI_ALARM_RED;
			qspan->span.alarms &= ~DAHDI_ALARM_YELLOW;
		} else {
			qspan->span.alarms &= ~DAHDI_ALARM_RED;
		}
		if (framer_checkbit(pmc, i, STAT0_AIS_HP)) {
			qspan->span.alarms |= DAHDI_ALARM_BLUE;
			qspan->span.alarms &= ~DAHDI_ALARM_YELLOW;
		} else {
			qspan->span.alarms &= ~DAHDI_ALARM_BLUE;
		}
		dahdi_alarm_notify(&qspan->span);

		DAHDI_COUNT(qspan->span, fascount, count.fas) =
			framer_read_pr(pmc, i, CNT_FAS_HP);

		reg = framer_read_pr(pmc, i, CNT_BPVHI_HP);
		reg <<= 8;
		reg |= framer_read_pr(pmc, i, CNT_BPVLO_HP);
		DAHDI_COUNT(qspan->span, bpvcount, count.bpv) = reg;

		reg = framer_read_pr(pmc, i, CNT_CRC4HI_HP);
		reg <<= 8;
		reg |= framer_read_pr(pmc, i, CNT_CRC4LO_HP);
		DAHDI_COUNT(qspan->span, crc4count, count.crc4) = reg;

		qspan->span.irqmisses = dpmc->irqmisses;

		if (framer_checkbit(pmc, i, STAT0_LOF_HP)) {
			qspan->frok = 0; /* Frame is bad */
		} else {
			qspan->frok++;
		}
		if (!qspan->frok || 8 == qspan->frok) {
			/* Reset equalizer if LOF detected and 1s after LOF finished. */
			reg = framer_read(pmc, i, 0x1, 0x1F);
			framer_write(pmc, i, 0x1, 0x1F, reg&(~(0x18))); /* switch of auto equalizer */
			framer_write(pmc, i, 0x1, 0x1F, reg|0x18); /* switch on auto equalizer */
		}
	}
}

void e12zap_clearstat(struct quasar_dpmc *dpmc) {
	struct quasar_pmc * pmc = dpmc->pvt;
	struct quasar_span *qspan;
	int i;

	for (i=0; i < dpmc->num_spans; ++i) {
		qspan = dpmc->spans[i];

		if (0 == qspan->clear_counters) continue; /* Already cleared */
		--qspan->clear_counters;
		if (0 != qspan->clear_counters) continue;
		module_dprintk(KERN_DEBUG, "e12zap_clearstat:"
			       "Clearing span %d error counters\n", qspan->span.spanno);
		framer_write_pr(pmc, i, CNT_FAS_HP, 0);
		framer_write_pr(pmc, i, CNT_BPVHI_HP, 0);
		framer_write_pr(pmc, i, CNT_BPVLO_HP, 0);
		framer_write_pr(pmc, i, CNT_CRC4HI_HP, 0);
		framer_write_pr(pmc, i, CNT_CRC4LO_HP, 0);
		dpmc->irqmisses = 0;
	}
}

static int e12zap_maint(struct dahdi_span *span, int cmd)
{
	/* Not implemented due to limited functionality. Use ecfg utility. */
	int res = 0;
	switch(cmd) {
	case DAHDI_MAINT_NONE:
		/* lloop=off, rloop=off */
	case DAHDI_MAINT_LOCALLOOP:
		/* lloop=on, rloop=? */
	case DAHDI_MAINT_REMOTELOOP:
		/* lloop=?, rloop=on */
	case DAHDI_MAINT_LOOPUP:
	case DAHDI_MAINT_LOOPDOWN:
	case DAHDI_MAINT_LOOPSTOP:
		span->mainttimer = span->maintstat = 0;
		module_printk(KERN_WARNING, "e21zap_maint: Use ecfg testing options.\n");
		res = -ENOSYS;
		break;
	default:
		module_printk(KERN_ERR, "e21zap_maint: Unknown maint command: %d\n", cmd);
		res = -ENOTTY;
		break;
	}
	return res;
}

/* trunk for CAS nibles... put CAS nibbles to appropriate position */
static int e12zap_rbsbits(struct dahdi_chan *chan, int bits)
{
	struct quasar_span *qspan = chan->pvt;
	struct quasar_dpmc *dpmc = qspan->pvt;
	struct quasar_pmc *pmc = dpmc->pvt;

	module_dprintk(KERN_DEBUG, "e12zap_rbsbits:"
		       "->%x on %s\n", bits, chan->name);

	if (chan->chanpos == 16)
		return 0;

	if(chan->chanpos < 16){
		framer_write(pmc, qspan->spannumber, 0x5, 0x10 + chan->chanpos, bits);
	}
	else {
		framer_write(pmc, qspan->spannumber, 0x6, 0x10 + chan->chanpos, bits);
	}
	return 0;
}

static int e12zap_chanconfig(struct dahdi_chan *chan, int sigtype)
{
	return 0;
}

static int e12zap_ioctl(struct dahdi_chan *chan, unsigned int cmd, unsigned long data)
{
	return -ENOTTY;
}

static int e12zap_close(struct dahdi_chan *chan)
{
	module_dprintk(KERN_DEBUG, "e12zap_close\n");
	module_put(THIS_MODULE);

	return 0;
}

static int e12zap_open(struct dahdi_chan *chan)
{
	module_dprintk(KERN_DEBUG, "e12zap_open\n");
	try_module_get(THIS_MODULE);

	return 0;
}

static int e12zap_shutdown(struct dahdi_span *span)
{
	int i;
	struct quasar_span *qspan = qspan_from_dspan(span);
	struct quasar_dpmc *dpmc = qspan->pvt;
	struct quasar_pmc *pmc = dpmc->pvt;

	module_dprintk(KERN_DEBUG, "e12zap_shutdown\n");

	if (!(span->flags & DAHDI_FLAG_RUNNING))
                return 0;

	span->flags &= ~DAHDI_FLAG_RUNNING;
	for (i=0; i < dpmc->num_spans; i++) {
		if (dpmc->spans[i]->span.flags & DAHDI_FLAG_RUNNING)
			return 0;
	}

	/* SPAN was running and it's the last. Lets stop the board. */

	/* disable DMA on receive and transmit */
	i = pmc->bar0[0x09];
	i &= ~(DMA_WENA|DMA_RENA);
	pmc->bar0[0x09] = i;

	/* cancel pending interrupts */
	pmc->bar0[0] = 0;

	/* disable DMA interrupt, mask */
	i = pmc->bar0[0x01];
	i &= ~(0x20000000);
	pmc->bar0[0x01] = i;

	return 0;
}

static int e12zap_startup(struct dahdi_span *span)
{
	int i;
	struct quasar_span *qspan = qspan_from_dspan(span);
	struct quasar_dpmc *dpmc = qspan->pvt;
	struct quasar_pmc * pmc = dpmc->pvt;
	int alreadyrunning = span->flags & DAHDI_FLAG_RUNNING;

	module_dprintk(KERN_DEBUG, "e12zap_startup: ");

	/* initialize the start value for the entire chunk of last ec buffer */
	for(i = 0; i < span->channels; i++) {
		memset(qspan->ec_chunk1[i],
			DAHDI_LIN2X(0,span->chans[i]),DAHDI_CHUNKSIZE);
		memset(qspan->ec_chunk2[i],
			DAHDI_LIN2X(0,span->chans[i]),DAHDI_CHUNKSIZE);
	}
	dprintk("", "Using ");

	if(span->lineconfig & DAHDI_CONFIG_HDB3) {
		framer_setbit(pmc, qspan->spannumber, CFG0_RCV_HDB3_HP);
		framer_setbit(pmc, qspan->spannumber, CFG0_XMIT_HDB3_HP);
		dprintk("", "HDB3/");
	} else {
		framer_clrbit(pmc, qspan->spannumber, CFG0_RCV_HDB3_HP);
                framer_clrbit(pmc, qspan->spannumber, CFG0_XMIT_HDB3_HP);
		dprintk("", "AMI/");
	}

	if(span->lineconfig & DAHDI_CONFIG_CCS) {
		framer_setbit(pmc, qspan->spannumber, CFG0_CAS_HP);
		dprintk("", "CCS/");
	} else {
		framer_clrbit(pmc, qspan->spannumber, CFG0_CAS_HP);
		dprintk("", "CAS/");
	}

	/* Basic CRC-4 synchronization processing. */
	framer_clrbit(pmc, qspan->spannumber, CFG3_CRC4_DIS_HP);
	/* Automatic CRC-interworking */
	framer_clrbit(pmc, qspan->spannumber, CFG3_CRC4_MANUAL_HP);

	/* Should I strict DAHDI's readme. See quasar/crc4.txt notes. */
	if (span->lineconfig & DAHDI_CONFIG_CRC4) {
		dprintk("", "CRC4 ");
	} else {
		dprintk("", "non-CRC4 ");
	}

	dprintk("", "on span %s, ", span->name);

	i = pmc->bar0[0x06]; /* get bs/ls */
	if (qspan->syncprior) {
		i &= ~(1<<qspan->spannumber);
		dprintk("", "can be the sync. source\n");
	} else {
		i |= (1<<qspan->spannumber);
		dprintk("", "can not be the sync. source\n");
	}
	pmc->bar0[0x06] = i; /* restore bs/ls */

	if (!alreadyrunning) {
		module_dprintk(KERN_DEBUG, "Enable DMA...\n");
                /* Enable receive/transmit DMA if not enabled. */
		i = pmc->bar0[0x09];
		i |= (DMA_WENA | DMA_RENA);
		pmc->bar0[0x09] = i;
		/* allow DMA interrupts, unmask, allow CASsig interrupt */
		i = pmc->bar0[0x01];
		i |= 0x20000000;
		pmc->bar0[0x01] = i;
		span->flags |= DAHDI_FLAG_RUNNING;
	}
	qspan->clear_counters = 3;

	return 0;
}

static int e12zap_spanconfig(struct dahdi_span *span, struct dahdi_lineconfig *lc)
{
	struct quasar_span *qspan = qspan_from_dspan(span);;

	module_dprintk(KERN_DEBUG, "e12zap_spanconfig:"
		       "Configuring span %d, span=%p\n", span->spanno, span);

	span->lineconfig = lc->lineconfig;
	span->txlevel = lc->lbo;
	span->rxlevel = 0;
	/* sync priority for SPAN within board (dahdi_cfg, system.conf).
	 1 = highest, 0 = freerun mode. */
	qspan->syncprior = lc->sync;

	/* If we're already running, then go ahead and apply the changes */
	if (span->flags & DAHDI_FLAG_RUNNING) {
		module_printk(KERN_INFO, "Starting span!\n");
		return e12zap_startup(span);
	}
	module_dprintk(KERN_DEBUG, "Done with spanconfig!\n");
	return 0;
}

#ifdef E12ZAP_TASKLET
static void e12zap_tasklet(unsigned long data)
{
	struct quasar_dpmc *dpmc = (struct quasar_dpmc *)data;
	dpmc->taskletrun++;
	if (dpmc->taskletpending) {
		dpmc->taskletexec++;
		e12zap_proc(dpmc);
	}
	dpmc->taskletpending = 0;
}
#endif

int e12zap_dacs(struct dahdi_chan *dst, struct dahdi_chan *src);
#ifdef HAVE_SPAN_OPS
static const struct dahdi_span_ops e12zap_span_ops = {
	.owner = THIS_MODULE,
	.spanconfig = e12zap_spanconfig,
	.chanconfig = e12zap_chanconfig,
	.startup = e12zap_startup,
	.shutdown = e12zap_shutdown,
	.rbsbits = e12zap_rbsbits,
	.maint = e12zap_maint,
	.open = e12zap_open,
	.close = e12zap_close,
	.ioctl = e12zap_ioctl,
	.dacs = e12zap_dacs,
};
void e12zap_setops(struct dahdi_span *span) {
	span->ops = &e12zap_span_ops;
}
#else //!HAVE_SPAN_OPS
void e12zap_setops(struct dahdi_span *span) {
	span->spanconfig = e12zap_spanconfig;
	span->chanconfig = e12zap_chanconfig;
	span->startup = e12zap_startup;
	span->shutdown = e12zap_shutdown;
	span->rbsbits = e12zap_rbsbits;
	span->maint = e12zap_maint;
	span->open = e12zap_open;
	span->close  = e12zap_close;
	span->ioctl = e12zap_ioctl;
	span->dacs = e12zap_dacs;
#ifdef HAVE_SPAN_OWNER
	span->owner = THIS_MODULE;
#endif //HAVE_SPAN_OWNER
}
#endif //HAVE_SPAN_OPS

int e12zap_register(struct quasar_pmc *pmc)
{
	struct pci_dev *pdev = pmc->pci_device;
	struct quasar_dpmc *dpmc;
	int spanid, chanid;

	pmc->priv_data = kmalloc(sizeof(struct quasar_dpmc), GFP_KERNEL);
	if (pmc->priv_data == NULL)
		return -ENOMEM;

	memset(pmc->priv_data, 0x0, sizeof(struct quasar_dpmc));
	dpmc = (struct quasar_dpmc *)pmc->priv_data;
	dpmc->pvt = pmc;

	/* QUASAR_BASE_SIZE*2(double buffer)*2(rx/tx) = 8192 */
	dpmc->writechunk  = (unsigned int *)pci_alloc_consistent(pdev, QUASAR_BASE_SIZE*2*2, &dpmc->writedma);
	if (dpmc->writechunk == NULL) {
		module_printk(KERN_INFO, "Unable to allocate DMA buffer\n");
		kfree(pmc->priv_data);
		return -ENOMEM;
	}

	/* Read is after the whole write piece (in words) */
	dpmc->readchunk = dpmc->writechunk + (QUASAR_BASE_SIZE*2/4);
	/* Same thing but in bytes...  */
	dpmc->readdma = dpmc->writedma + (QUASAR_BASE_SIZE * 2);

	dpmc->num_spans = pmc->e1_amount + 1;
	if (dpmc->num_spans == 2)
		dpmc->qn_span = 2;
	else
		dpmc->qn_span = 4;

	/* Initialize Write/Buffers to all blank data */
	memset((void *)dpmc->writechunk,0x00, QUASAR_BASE_SIZE*2);
	memset((void *)dpmc->readchunk,0xff, QUASAR_BASE_SIZE*2);

	pci_set_master(pdev);

	/* Allocate memory for the spans */
	for (spanid = 0; spanid < dpmc->num_spans; spanid++) {
		dpmc->spans[spanid] = kmalloc(sizeof(struct quasar_span), GFP_KERNEL);
		if (dpmc->spans[spanid]) {
			memset(dpmc->spans[spanid], 0, sizeof(struct quasar_span));
		} else {
			goto unreg_all;
		}
	}

	/* Initialize spans */
	for (spanid = 0; spanid < dpmc->num_spans; spanid++) {
		struct quasar_span *qspan = dpmc->spans[spanid];
		struct dahdi_span *span = &(qspan->span);
		qspan->spannumber = spanid;
		qspan->pvt = dpmc;
		span->channels = 31;
		span->spantype = "E1";
		span->manufacturer = "Parabel ltd.";
		sprintf(span->name,
			"Quasar_%d_%d", pmc->dev_number, spanid+1);
		sprintf(span->desc, "Parabel Quasar %d Span %d",
			pmc->dev_number, spanid+1);
		snprintf(span->location, sizeof(span->location) - 1,
			 "PCI Bus %02d Slot %02d", pdev->bus->number,
			 PCI_SLOT(pdev->devfn) + 1);
		snprintf(span->devicetype, sizeof(span->devicetype) - 1,
			 "Quasar-%d PCI/PCIe/PMC, hw:0x%X",
			 dpmc->num_spans, pmc->bar0[0x0D]);
		span->chans = dpmc->spans[spanid]->chans;
		span->flags = DAHDI_FLAG_RBS;
		span->linecompat = DAHDI_CONFIG_HDB3 | DAHDI_CONFIG_CCS | DAHDI_CONFIG_CRC4 | DAHDI_CONFIG_AMI;
		span->deflaw = DAHDI_LAW_ALAW;

		e12zap_setops (&dpmc->spans[spanid]->span);

		init_waitqueue_head(&(span->maintq));

		for (chanid=0; chanid < span->channels; chanid++) {
                        struct dahdi_chan *chan;
			chan = dpmc->spans[spanid]->chans[chanid] = &dpmc->spans[spanid]->_chans[chanid];
			sprintf(chan->name, "Quasar_%d_%d_%d", pmc->dev_number, spanid + 1, chanid + 1);
			chan->sigcap = DAHDI_SIG_EM | DAHDI_SIG_CLEAR | DAHDI_SIG_EM_E1 |
				DAHDI_SIG_FXSLS | DAHDI_SIG_FXSGS |
				DAHDI_SIG_FXSKS | DAHDI_SIG_FXOLS |
				DAHDI_SIG_FXOGS | DAHDI_SIG_FXOKS |
				DAHDI_SIG_CAS | DAHDI_SIG_SF | DAHDI_SIG_DACS_RBS;
			chan->pvt = dpmc->spans[spanid];
			chan->chanpos = chanid + 1; /* timeslot number */
		}

		if (dahdi_register(span, 0)) {
			module_printk(KERN_ERR, "Unable to register span Quasar_%d_%d with dahdi!\n", pmc->dev_number, spanid+1);
			for (spanid=(spanid-1); spanid >= 0; spanid--) {
				module_printk(KERN_INFO, "Unregister span %d\n", spanid);
				dahdi_unregister(&dpmc->spans[spanid]->span);
			}
			goto unreg_all;
		}
	}
#ifdef E12ZAP_TASKLET
	tasklet_init(&dpmc->e12zap_tlet, e12zap_tasklet, (unsigned long)dpmc);
#endif
	/* Initialize hardware */
	pmc->bar0[0x0a] = dpmc->readdma;
	pmc->bar0[0x0c] = dpmc->writedma;
	pmc->bar0[0x09] = DMA_IFQ0 | DMA_SIZE0;
	dpmc->registered = 1;

	return 0;

	/* TODO: no goto ! */
unreg_all:
	module_printk(KERN_ERR, "e12zap_register: Not enough free resources!\n");
	/* free spans */
	for (spanid = 0; spanid < dpmc->num_spans; spanid++) {
		if (dpmc->spans[spanid]) {
			kfree(dpmc->spans[spanid]);
			dpmc->spans[spanid] = NULL;
		}
	}
	pci_free_consistent(pdev, QUASAR_BASE_SIZE*2*2, (void *) dpmc->writechunk, dpmc->writedma);
	kfree(pmc->priv_data);

	return -ENOMEM;
}

int e12zap_unregister(struct quasar_pmc *pmc) {
	struct pci_dev *pdev = pmc->pci_device;
	struct quasar_dpmc *dpmc;
	int spanid, tmp;

	dpmc = (struct quasar_dpmc *)pmc->priv_data;

	if (dpmc) {
		if (dpmc->registered) {
			/* detach from dahdi */
			for(spanid=0; spanid < dpmc->num_spans; spanid++) {
				dahdi_unregister(&dpmc->spans[spanid]->span);
				kfree(dpmc->spans[spanid]);
			}
			pci_free_consistent(pdev, QUASAR_BASE_SIZE*2*2, (void *) dpmc->writechunk, dpmc->writedma);
			kfree(pmc->priv_data);
			/* deinitialize hardware */

			/* disable DMA on recv and transmit */
			tmp = pmc->bar0[0x09];
			tmp &= ~(DMA_WENA|DMA_RENA);
			pmc->bar0[0x09] = tmp;

			/* disable DMA interrupt, mask */
			tmp = pmc->bar0[0x01];
			tmp &= ~(0x02000000);
			pmc->bar0[0x01] = tmp;
		}
	}
	return 0;
}

void e12zap_proc(struct quasar_dpmc *dpmc) {
	int spanid, chanid;

	for (spanid=0; spanid < dpmc->num_spans; spanid++) {
                struct quasar_span *qspan = dpmc->spans[spanid];
		if (qspan->span.flags &  DAHDI_FLAG_RUNNING) {
			/*Echo cancelation*/
			for(chanid=0; chanid < qspan->span.channels; chanid++) {
				dahdi_ec_chunk(qspan->span.chans[chanid],
					       qspan->span.chans[chanid]->readchunk,
					       qspan->ec_chunk2[chanid]);
				memcpy(qspan->ec_chunk2[chanid], qspan->ec_chunk1[chanid],
				       DAHDI_CHUNKSIZE);
				memcpy(qspan->ec_chunk1[chanid], qspan->span.chans[chanid]->writechunk,
				       DAHDI_CHUNKSIZE);
			}
			dahdi_receive(&qspan->span);
		}
	}
	for (spanid=0; spanid < dpmc->num_spans; spanid++) {
		if (dpmc->spans[spanid]->span.flags &  DAHDI_FLAG_RUNNING) {
			dahdi_transmit(&dpmc->spans[spanid]->span);
		}
	}
}

/*
 1. Free and fill system DMA buffers prepare to call dahdi_receive and dahdi_transmit.
 2. Check and renew CAS signalization.
 3. Check loosing of sync.
 4. Perform echo cancelation on RX chanks.
 5. Call dahdi_receive and dahdi_transmit. //process independently
 */
/* TODO: cleanup code. */
int e12zap_delivery(struct quasar_pmc *pmc) {
	struct quasar_dpmc *dpmc = pmc->priv_data;
	unsigned int keep, chunk, spanid, chanid, tmp1, tmp2, shift;

	if (!pmc->priv_data) {
		return -1;
	}

#ifdef E12ZAP_TASKLET
	if (!dpmc->taskletpending) {
		dpmc->taskletpending = 1;
		dpmc->taskletsched++;
		tasklet_schedule(&dpmc->e12zap_tlet);
	} else {
		module_dprintk(KERN_DEBUG, "Miss tasklet. Shed:%d Exec:%d Run:%d, TX_Err: %d\n",
			       dpmc->taskletsched, dpmc->taskletexec, dpmc->taskletrun, dpmc->txerrors++);
		dpmc->txerrors++;
	}
#endif
	/* check DMA overflow counter */
	keep = (pmc->bar0[0x09] & 0xff000000) >> 24;

	if (keep != dpmc->ovfl_counter)
		module_dprintk(KERN_DEBUG, "DMA OVFL:%d\n", dpmc->ovfl_counter);
	dpmc->ovfl_counter = keep;

	keep = pmc->bar0[0x0b]; /* get relative addres of current dma operating position */

	if (keep < (DAHDI_CHUNKSIZE * 256)) {
		/* we are on the first part of buffer, serve second part of the buffer */
		if (dpmc->check == 1) {
			++dpmc->irqmisses;
			module_dprintk(KERN_DEBUG, "Miss chunk in 1st part.\n");
		}

		dpmc->check = 1;
		//	    if (keep > 0x80) printk("e12zap: Possible data loss=0x%x\n", keep);
		for (chunk = 0; chunk < DAHDI_CHUNKSIZE; chunk++) {
			for (spanid=0; spanid < dpmc->qn_span; spanid ++) { /* span */
				for (chanid=0; chanid<16; chanid++) { /* chan */
					shift = chanid + 16 * spanid +  64 * chunk + DAHDI_CHUNKSIZE * 64;
					tmp2 = dpmc->readchunk[shift];
					tmp1 = 0;
					if (chanid==0) { /* for 1 timeslot */
						/* transmit data from dahdi 1ts */
						if (dpmc->num_spans > 4) {
							dpmc->spans[4+spanid]->span.chans[0]->readchunk[chunk] = 0xff & (tmp2 >> 16);
							tmp1 =  (dpmc->spans[4+spanid]->span.chans[0]->writechunk[chunk] << 16);
						}
						dpmc->spans[spanid]->span.chans[0]->readchunk[chunk] = 0xff & tmp2;
						tmp1 |=  dpmc->spans[spanid]->span.chans[0]->writechunk[chunk];

					} else {
						/* transmit data from dahdi 2-31 ts */
						if (dpmc->num_spans > 4) {
							dpmc->spans[4+spanid]->span.chans[chanid*2-1]->readchunk[chunk] = tmp2 >> 24;
							dpmc->spans[4+spanid]->span.chans[chanid*2]->readchunk[chunk] = 0xff & (tmp2 >> 16);
							tmp1 =  (dpmc->spans[4+spanid]->span.chans[chanid*2-1]->writechunk[chunk] << 24);
							tmp1 |= (dpmc->spans[4+spanid]->span.chans[chanid*2]->writechunk[chunk] << 16);

						}
						dpmc->spans[spanid]->span.chans[chanid*2-1]->readchunk[chunk] = 0xff & (tmp2 >> 8);
						dpmc->spans[spanid]->span.chans[chanid*2]->readchunk[chunk] = 0xff & tmp2;
						tmp1 |= (dpmc->spans[spanid]->span.chans[chanid*2-1]->writechunk[chunk] << 8);
						tmp1 |= (dpmc->spans[spanid]->span.chans[chanid*2]->writechunk[chunk]);
					}
					dpmc->writechunk[shift] = tmp1;
				}
			}
		}

	} else {
		/* we are in second part of the buffer, serve first part of the buffer */

		if (dpmc->check == 2) {
			++dpmc->irqmisses;
			module_dprintk(KERN_DEBUG, "Miss chunk in 2nd part.\n");
		}
		dpmc->check = 2;
		//	    if (keep > 0x880) printk("e12zap: Possible data loss=0x%x\n", keep);
		for (chunk = 0; chunk < DAHDI_CHUNKSIZE; chunk++) {
			for(spanid=0; spanid < dpmc->qn_span; spanid ++) { /* span */
				for (chanid=0; chanid < 16 ;chanid++) { /* chan */
					shift = chanid + 16 * spanid +  64 * chunk;
					tmp2 = dpmc->readchunk[shift];
					tmp1 = 0;
					if (chanid==0) { /* for 1 timeslot */
						/* transmit data from dahdi 1ts */
						if (dpmc->num_spans > 4) {
							dpmc->spans[4+spanid]->span.chans[0]->readchunk[chunk] = 0xff & (tmp2 >> 16);
							tmp1 =  (dpmc->spans[4+spanid]->span.chans[0]->writechunk[chunk] << 16);
						}
						dpmc->spans[spanid]->span.chans[0]->readchunk[chunk] = 0xff & tmp2;
						tmp1 |=  dpmc->spans[spanid]->span.chans[0]->writechunk[chunk];

					} else {
						/* transmit data from dahdi 2-31 ts */
						if (dpmc->num_spans > 4) {
							dpmc->spans[4+spanid]->span.chans[chanid*2-1]->readchunk[chunk] = tmp2 >> 24;
							dpmc->spans[4+spanid]->span.chans[chanid*2]->readchunk[chunk] = 0xff & (tmp2 >> 16);

							tmp1 =  (dpmc->spans[4+spanid]->span.chans[chanid*2-1]->writechunk[chunk] << 24);
							tmp1 |= (dpmc->spans[4+spanid]->span.chans[chanid*2]->writechunk[chunk] << 16);
						}
						dpmc->spans[spanid]->span.chans[chanid*2-1]->readchunk[chunk] = 0xff & (tmp2 >> 8);
						dpmc->spans[spanid]->span.chans[chanid*2]->readchunk[chunk] = 0xff & tmp2;

						tmp1 |= (dpmc->spans[spanid]->span.chans[chanid*2-1]->writechunk[chunk] << 8);
						tmp1 |= (dpmc->spans[spanid]->span.chans[chanid*2]->writechunk[chunk]);
					}
					dpmc->writechunk[shift] = tmp1;
				}
			}
		}
	}

	/* Check signalisation */
	for (spanid=0; spanid < dpmc->num_spans; spanid++) {
		if (dpmc->spans[spanid]->span.flags & DAHDI_FLAG_RUNNING) {
			/* read interrupt status */
			if (framer_read(pmc, spanid, 0x4, 0x1e) & 0x01) {
				for(chanid=1; chanid < 16; chanid++) {
					/* chans < 16                            dahdi channels [1:15]  - chans [0:14] */
					tmp1 = framer_read(pmc, spanid, 0x9, 0x10 + chanid) & 0x0f;
					if (tmp1 != dpmc->spans[spanid]->span.chans[chanid-1]->rxsig) {
						if(!(dpmc->spans[spanid]->span.chans[chanid-1]->sig & DAHDI_SIG_CLEAR)) {
							dahdi_rbsbits(dpmc->spans[spanid]->span.chans[chanid-1], tmp1);
						}
					}
				}

				for(chanid=1; chanid < 16; chanid++) {
					/* chans > 16                            dahdi channels [17:31] -chans [16:30] */
					tmp1 = framer_read(pmc, spanid, 0xa, 0x10 + chanid) & 0x0f;
					if (tmp1 != dpmc->spans[spanid]->span.chans[15+chanid]->rxsig) {
						if(!(dpmc->spans[spanid]->span.chans[15+chanid]->sig & DAHDI_SIG_CLEAR)) {
							dahdi_rbsbits(dpmc->spans[spanid]->span.chans[15+chanid], tmp1);
						}
					}
				}
			}
		}
	}

#ifndef E12ZAP_TASKLET
	e12zap_proc(dpmc);
#endif
	++dpmc->intcount;
	if (0 == (dpmc->intcount & 127)) {
		e12zap_check_alarm (dpmc);
		e12zap_find_master (dpmc);
		e12zap_clearstat (dpmc);
	}

	return 0;
}

int e12zap_dacs (struct dahdi_chan *dst, struct dahdi_chan *src) {
	struct quasar_span *qspan_dst = dst->pvt;
	struct quasar_dpmc *dpmc = qspan_dst->pvt;
	struct quasar_pmc *pmc = dpmc->pvt;

	module_dprintk(KERN_DEBUG, "e12zap_dacs:\n");

	if (!dst)
		return -1;

	if (src) {
		struct quasar_span *qspan_src = src->pvt;
		if(qspan_dst->pvt != qspan_src->pvt) {
			/* trying to cross-connect timeslots from different boards */
			module_dprintk(KERN_DEBUG, "Can't switch timeslots from different boards, using soft switch\n");

			if (qspan_dst->spannumber < 4) {
				pmc->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(4, (dst->chanpos + 32 * qspan_dst->spannumber));
			} else {
				pmc->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(5, (dst->chanpos + 32 * (qspan_dst->spannumber - 4)));
			}
			return -1;
		}
		module_dprintk(KERN_DEBUG, "Switching %d:%d(in)/%d:%d(out).\n", qspan_dst->spannumber, dst->chanpos, qspan_src->spannumber, src->chanpos);
		/* cross-connecting
		 4 -> 6
		 5 -> 7 */
		if (qspan_src->spannumber < 4) {
			pmc->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(6, (src->chanpos + 32 * qspan_src->spannumber));
		} else {
			pmc->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(7, (src->chanpos + 32 * (qspan_src->spannumber - 4)));
		}
	} else {
		/* return to default crossing */
		module_dprintk(KERN_DEBUG, "Return to default crossing on %d:%d\n", qspan_dst->spannumber, dst->chanpos);

		if (qspan_dst->spannumber < 4) {
			pmc->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(4, (dst->chanpos + 32 * qspan_dst->spannumber));
		} else {
			pmc->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(5, (dst->chanpos + 32 * (qspan_dst->spannumber - 4)));
		}
	}
	return 0;
}
